function im_bin = imphansalkar(im_input,w_radius, p, r_factor, restrict)
% imphansalkar: a function for applying an adaptive Phansalkar threshold to 
% an image
%   
%   'im_bin'    - binary image with the same dimensions as im_input.
%   'im_input'  - greyscale 2D or 3D image to be thresholded
%   'w_radius'  - radius for calculating the mean and std for the threshold
%   'p'         - free parameter, must be greater than 0 (default 2)
%   'r_factor'  - free parameter, must be greater than or equal to 1 (default 1)
%   'restrict'  - method of restriction of parameter selection, currently
%   only available option is 'otsu'.

[height,width,depth] = size(im_input);
w_size = 2 * w_radius + 1;
if numel(w_size) == 1
    w_size = [min(w_size,height),min(w_size,width),min(w_size,depth)];
end
w_volume = w_size(1) * w_size(2) * w_size(3);

if exist('p','var')
    if (numel(p) ~= 1) || ~isnumeric(p) || (p <= 0)
        p = 2;
    end
else
    p = 2;
end

if exist('r_factor','var')
    if (numel(r_factor) ~= 1) || ~isnumeric(r_factor) || (r_factor < 1)
        r_factor = 1;
    end
else
    r_factor = 1;
end

% normalize image to [0,1]
min_im = min(im_input(:));
im_input = im_input - min_im;
max_im = max(im_input(:));
im_input = im_input / max_im;

% generate local mean and standard deviation images
im_mean = movmean(im_input,w_size(1),1);
for d = 2:3
    im_mean = movmean(im_mean,w_size(d),d);
end
im_mean2 = im_input.^2;
for d = 1:3
    im_mean2 = movsum(im_mean2,w_size(d),d);
end
im_mean2 = im_mean2 / (w_volume - 1);
im_std = real(sqrt(im_mean2 - im_mean.^2 * w_volume / (w_volume - 1)));

% approximate local mean and standard deviation values for foreground and
% background pixels.
restrict_mode = false;
if exist('restrict','var')
    restrict_mode = true;
    if strcmp(restrict, 'otsu')
        % restrict parameter estimation inputs to pixels with local mean
        % above the Otsu threshold
        im_restrict = imbinarize(im_mean);
        % dilate to compensate for reduced volume induced by mean filter
        im_dist = bwdist(im_restrict);
        im_restrict = 1 * (im_dist < w_radius(1));
    else
        im_restrict = ones(height,width,depth);
    end
    sigma_mu = std(im_mean(im_restrict == 1));
    mu_mu = mean(im_mean(im_restrict == 1));
    std_FG = max(im_std(im_restrict == 1));
else
    sigma_mu = std(im_mean(:));
    mu_mu = mean(im_mean(:));
    std_FG = max(im_std(:));
end
mean_FG = mu_mu + 2 * sigma_mu;
mean_BG = mu_mu + sigma_mu;

% select q such that the exponential term in the threshold has half-life 
% 2 * sigma_mu
q = log(2) / (2 * sigma_mu);
% select remaining parameter values such that 
% a) pixels with 
% local mean > mean_FG and local std = std_FG have
% threshold < local mean
% b) pixels with
% local mean < mean_BG and local std = 0 have
% threshold > local mean
r = r_factor * std_FG / (1 - exp(-q * (mean_FG - mean_BG)));
k = p * (exp(-q * mean_BG) + exp(-q * mean_FG) / (1 - std_FG / r)) / 2;

% threshold formula from Phansalkar 2011
im_threshold = im_mean .* (1 + p * exp(-q * im_mean) + k * (im_std / r - 1));
% apply threshold
im_bin = zeros(height,width,depth);
if restrict_mode
    im_bin(im_input > im_threshold) = im_restrict(im_input > im_threshold);
else
    im_bin(im_input > im_threshold) = 1;
end
end

