function A = pgmread3(path)
% reads a 3D pgm file and returns as a 3D image A
%
% pgm files read by this function have the header
% 'P5
% width height depth
% max_value'
% where
% P5 describes the pgm file
% width, height, and depth are image size parameters
% max_value is the maximum pixel value of the image (currently only 255
% supported).
%
%   'A'     -   output image from the file, with dimensions specified in
%               the file header
%   'path'  -   path to the image file

% parameter to limit header reading (error if header is longer than this)
max_header_length = 256;

% check file is readable
assert(exist(path,'file') ~= 0,'Error: could not find image file.');

% open file as char array, producing double vector (default conversion)
fid = fopen(path);
raw_input = fread(fid);
fclose(fid);

% find dimensions (second line) and end of header (fourth line)
line_number = 1;
% dimensions metadata
dim_char = char(ones(1,max_header_length)*48);
dim_pos = 1;
% maximum value metadata (not currently used)
max_char = char(ones(1,max_header_length)*48);
max_pos = 1;

data_start = 1;
for h_i = 1:max_header_length
    if char(raw_input(h_i)) == newline
        line_number = line_number + 1;
    end
    % dimension line
    if line_number == 2
        dim_char(dim_pos) = char(raw_input(h_i));
        dim_pos = dim_pos + 1;
    % max value line
    elseif line_number == 3
        max_char(max_pos) = char(raw_input(h_i));
        max_pos = max_pos + 1;
    % data start line
    elseif line_number > 3
        data_start = h_i + 1;
        break;
    end
end
% trim char arrays
dim_char(dim_pos:end) = [];
dim_char(1) = [];
max_char(max_pos:end) = [];
max_char(1) = [];

% check file format is correct
assert(h_i < max_header_length,'Error: header format not supported.');
im_size = str2num(dim_char);
im_max = str2num(max_char);

A = zeros(im_size);

A(:) = raw_input(data_start:end);
% raw values stored as x + y * w + z * w * h, so x and y need to flip
A = permute(A,[2,1,3]);
